@@ -1,3 +1,6 @@
+# pylint:  disable=too-many-lines
+# (Yes, it has a point!)
+
 from __future__ import division, absolute_import, print_function
 
 __copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"
@@ -25,21 +28,134 @@ THE SOFTWARE.
 
 import operator
 import sys
-
-from pytools.decorator import decorator
-import six
-from six.moves import range, zip, intern, input
+import logging
 from functools import reduce
 
-try:
-    decorator_module = __import__("decorator", level=0)
-except TypeError:
-    # this must be Python 2.4
-    my_decorator = decorator
-except ImportError:
-    my_decorator = decorator
-else:
-    my_decorator = decorator_module.decorator
+import six
+from six.moves import range, zip, intern, input
+
+
+decorator_module = __import__("decorator", level=0)
+my_decorator = decorator_module.decorator
+
+__doc__ = """
+A Collection of Utilities
+=========================
+
+Math
+----
+
+.. autofunction:: levi_civita
+.. autofunction:: perm
+.. autofunction:: comb
+
+Assertive accessors
+-------------------
+
+.. autofunction:: one
+.. autofunction:: is_single_valued
+.. autofunction:: all_roughly_equal
+.. autofunction:: single_valued
+
+Memoization
+-----------
+
+.. autofunction:: memoize
+.. autofunction:: memoize_on_first_arg
+.. autofunction:: memoize_method
+.. autofunction:: memoize_method_with_uncached
+.. autofunction:: memoize_in
+
+Argmin/max
+----------
+
+.. autofunction:: argmin2
+.. autofunction:: argmax2
+.. autofunction:: argmin
+.. autofunction:: argmax
+
+Cartesian products
+------------------
+.. autofunction:: cartesian_product
+.. autofunction:: distinct_pairs
+
+Permutations, Tuples, Integer sequences
+---------------------------------------
+
+.. autofunction:: wandering_element
+.. autofunction:: indices_in_shape
+.. autofunction:: generate_nonnegative_integer_tuples_below
+.. autofunction:: generate_nonnegative_integer_tuples_summing_to_at_most
+.. autofunction:: generate_all_nonnegative_integer_tuples
+.. autofunction:: generate_all_integer_tuples_below
+.. autofunction:: generate_all_integer_tuples
+.. autofunction:: generate_permutations
+.. autofunction:: generate_unique_permutations
+
+Graph Algorithms
+----------------
+
+.. autofunction:: a_star
+
+Formatting
+----------
+
+.. autoclass:: Table
+.. autofunction:: string_histogram
+.. autofunction:: word_wrap
+
+Debugging
+---------
+
+.. autofunction:: typedump
+.. autofunction:: invoke_editor
+
+Progress bars
+-------------
+
+.. autoclass:: ProgressBar
+
+Name generation
+---------------
+
+.. autofunction:: generate_unique_names
+.. autofunction:: generate_numbered_unique_names
+.. autoclass:: UniqueNameGenerator
+
+Functions for dealing with (large) auxiliary files
+--------------------------------------------------
+
+.. autofunction:: download_from_web_if_not_present
+
+Helpers for :mod:`numpy`
+------------------------
+
+.. autofunction:: reshaped_view
+
+
+Timing data
+-----------
+
+.. data:: SUPPORTS_PROCESS_TIME
+
+   A :class:`bool` indicating whether :class:`ProcessTimer` measures elapsed
+   process time (available on Python 3.3+).
+
+.. autoclass:: ProcessTimer
+
+Log utilities
+-------------
+
+.. autoclass:: ProcessLogger
+.. autoclass:: DebugProcessLogger
+.. autoclass:: log_process
+
+Sorting in natural order
+------------------------
+
+.. autofunction:: natorder
+.. autofunction:: natsorted
+"""
 
 
 # {{{ math --------------------------------------------------------------------
@@ -127,9 +243,12 @@ class RecordWithoutPickling(object):
 
     __slots__ = []
 
-    def __init__(self, valuedict=None, exclude=["self"], **kwargs):
+    def __init__(self, valuedict=None, exclude=None, **kwargs):
         assert self.__class__ is not Record
 
+        if exclude is None:
+            exclude = ["self"]
+
         try:
             fields = self.__class__.fields
         except AttributeError:
@@ -170,6 +289,13 @@ class RecordWithoutPickling(object):
 
         fields.update(new_fields)
 
+    def __getattr__(self, name):
+        # This method is implemented to avoid pylint 'no-member' errors for
+        # attribute access.
+        raise AttributeError(
+                "'%s' object has no attribute '%s'" % (
+                    self.__class__.__name__, name))
+
 
 class Record(RecordWithoutPickling):
     __slots__ = []
@@ -200,12 +326,18 @@ class Record(RecordWithoutPickling):
 
 class ImmutableRecordWithoutPickling(RecordWithoutPickling):
     "Hashable record. Does not explicitly enforce immutability."
+    def __init__(self, *args, **kwargs):
+        RecordWithoutPickling.__init__(self, *args, **kwargs)
+        self._cached_hash = None
 
     def __hash__(self):
-        return hash(
+        if self._cached_hash is None:
+            self._cached_hash = hash(
                 (type(self),) + tuple(getattr(self, field)
                     for field in self.__class__.fields))
 
+        return self._cached_hash
+
 
 class ImmutableRecord(ImmutableRecordWithoutPickling, Record):
     pass
@@ -226,39 +358,6 @@ class Reference(object):
         self.value = value
 
 
-# {{{ dictionary with default
-
-class DictionaryWithDefault(object):
-    def __init__(self, default_value_generator, start={}):
-        self._Dictionary = dict(start)
-        self._DefaultGenerator = default_value_generator
-
-    def __getitem__(self, index):
-        try:
-            return self._Dictionary[index]
-        except KeyError:
-            value = self._DefaultGenerator(index)
-            self._Dictionary[index] = value
-            return value
-
-    def __setitem__(self, index, value):
-        self._Dictionary[index] = value
-
-    def __contains__(self, item):
-        return True
-
-    def iterkeys(self):
-        return six.iterkeys(self._Dictionary)
-
-    def __iter__(self):
-        return self._Dictionary.__iter__()
-
-    def iteritems(self):
-        return six.iteritems(self._Dictionary)
-
-# }}}
-
-
 class FakeList(object):
     def __init__(self, f, length):
         self._Length = length
@@ -278,7 +377,10 @@ class FakeList(object):
 # {{{ dependent dictionary ----------------------------------------------------
 
 class DependentDictionary(object):
-    def __init__(self, f, start={}):
+    def __init__(self, f, start=None):
+        if start is None:
+            start = {}
+
         self._Function = f
         self._Dictionary = start.copy()
 
@@ -287,7 +389,7 @@ class DependentDictionary(object):
 
     def __contains__(self, key):
         try:
-            self[key]
+            self[key]  # pylint: disable=pointless-statement
             return True
         except KeyError:
             return False
@@ -421,15 +523,15 @@ def memoize(*args, **kwargs):
             # http://www.phyast.pitt.edu/~micheles/python/
             key = key_func(*args, **kwargs)
             try:
-                return func._memoize_dic[key]
+                return func._memoize_dic[key]  # pylint: disable=protected-access
             except AttributeError:
                 # _memoize_dic doesn't exist yet.
                 result = func(*args, **kwargs)
-                func._memoize_dic = {key: result}
+                func._memoize_dic = {key: result}  # pylint: disable=protected-access
                 return result
             except KeyError:
                 result = func(*args, **kwargs)
-                func._memoize_dic[key] = result
+                func._memoize_dic[key] = result  # pylint: disable=protected-access
                 return result
     else:
         @my_decorator
@@ -437,15 +539,15 @@ def memoize(*args, **kwargs):
             # by Michele Simionato
             # http://www.phyast.pitt.edu/~micheles/python/
             try:
-                return func._memoize_dic[args]
+                return func._memoize_dic[args]  # pylint: disable=protected-access
             except AttributeError:
                 # _memoize_dic doesn't exist yet.
                 result = func(*args)
-                func._memoize_dic = {args: result}
+                func._memoize_dic = {args: result}  # pylint:disable=protected-access
                 return result
             except KeyError:
                 result = func(*args)
-                func._memoize_dic[args] = result
+                func._memoize_dic[args] = result  # pylint: disable=protected-access
                 return result
     if not args:
         return _deco
@@ -496,10 +598,9 @@ def memoize_on_first_arg(function, cache_dict_name=None):
     def clear_cache(obj):
         delattr(obj, cache_dict_name)
 
-    if sys.version_info >= (2, 5):
-        from functools import update_wrapper
-        new_wrapper = update_wrapper(wrapper, function)
-        new_wrapper.clear_cache = clear_cache
+    from functools import update_wrapper
+    new_wrapper = update_wrapper(wrapper, function)
+    new_wrapper.clear_cache = clear_cache
 
     return new_wrapper
 
@@ -514,13 +615,18 @@ def memoize_method(method):
     return memoize_on_first_arg(method, intern("_memoize_dic_"+method.__name__))
 
 
-def memoize_method_with_uncached(uncached_args=[], uncached_kwargs=set()):
+def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
     """Supports cache deletion via ``method_name.clear_cache(self)``.
 
     :arg uncached_args: a list of argument numbers
         (0-based, not counting 'self' argument)
     """
 
+    if uncached_args is None:
+        uncached_args = []
+    if uncached_kwargs is None:
+        uncached_kwargs = set()
+
     # delete starting from the end
     uncached_args = sorted(uncached_args, reverse=True)
     uncached_kwargs = list(uncached_kwargs)
@@ -674,7 +780,7 @@ def monkeypatch_method(cls):
     return decorator
 
 
-def monkeypatch_class(name, bases, namespace):
+def monkeypatch_class(_name, bases, namespace):
     # from GvR, http://mail.python.org/pipermail/python-dev/2008-January/076194.html
 
     assert len(bases) == 1, "Exactly one base class required"
@@ -720,13 +826,13 @@ def len_iterable(iterable):
     return sum(1 for i in iterable)
 
 
-def flatten(list):
+def flatten(iterable):
     """For an iterable of sub-iterables, generate each member of each
     sub-iterable in turn, i.e. a flattened version of that super-iterable.
 
     Example: Turn [[a,b,c],[d,e,f]] into [a,b,c,d,e,f].
     """
-    for sublist in list:
+    for sublist in iterable:
         for j in sublist:
             yield j
 
@@ -737,7 +843,7 @@ def general_sum(sequence):
 
 def linear_combination(coefficients, vectors):
     result = coefficients[0] * vectors[0]
-    for c, v in zip(coefficients, vectors)[1:]:
+    for c, v in zip(coefficients[1:], vectors[1:]):
         result += c*v
     return result
 
@@ -750,8 +856,8 @@ def common_prefix(iterable, empty=None):
         return empty
 
     for v in it:
-        for j in range(len(pfx)):
-            if pfx[j] != v[j]:
+        for j, pfx_j in enumerate(pfx):
+            if pfx_j != v[j]:
                 pfx = pfx[:j]
                 if j == 0:
                     return pfx
@@ -760,14 +866,14 @@ def common_prefix(iterable, empty=None):
     return pfx
 
 
-def decorate(function, list):
-    return [(x, function(x)) for x in list]
+def decorate(function, iterable):
+    return [(x, function(x)) for x in iterable]
 
 
-def partition(criterion, list):
+def partition(criterion, iterable):
     part_true = []
     part_false = []
-    for i in list:
+    for i in iterable:
         if criterion(i):
             part_true.append(i)
         else:
@@ -791,8 +897,8 @@ def product(iterable):
     return reduce(mul, iterable, 1)
 
 
-all = six.moves.builtins.all
-any = six.moves.builtins.any
+all = six.moves.builtins.all  # pylint: disable=redefined-builtin
+any = six.moves.builtins.any  # pylint: disable=redefined-builtin
 
 
 def reverse_dictionary(the_dict):
@@ -846,7 +952,7 @@ def find_max_where(predicate, prec=1e-5, initial_guess=1, fail_bound=1e38):
 
     # {{{ establish bracket
 
-    mag = 1
+    mag = initial_guess
 
     if predicate(mag):
         mag *= 2
@@ -887,8 +993,8 @@ def find_max_where(predicate, prec=1e-5, initial_guess=1, fail_bound=1e38):
             lower_true = mid
         else:
             upper_false = mid
-    else:
-        return lower_true
+
+    return lower_true
 
     # }}}
 
@@ -945,10 +1051,15 @@ def argmax(iterable):
 
 # {{{ cartesian products etc.
 
-def cartesian_product(list1, list2):
-    for i in list1:
-        for j in list2:
-            yield (i, j)
+def cartesian_product(*args):
+    if len(args) == 1:
+        for arg in args[0]:
+            yield (arg,)
+        return
+    first = args[:-1]
+    for prod in cartesian_product(*first):
+        for i in args[-1]:
+            yield prod + (i,)
 
 
 def distinct_pairs(list1, list2):
@@ -979,16 +1090,16 @@ def average(iterable):
     it = iterable.__iter__()
 
     try:
-        sum = next(it)
+        s = next(it)
         count = 1
     except StopIteration:
         raise ValueError("empty average")
 
     for value in it:
-        sum = sum + value
+        s = s + value
         count += 1
 
-    return sum/count
+    return s/count
 
 
 class VarianceAggregator:
@@ -1005,9 +1116,9 @@ class VarianceAggregator:
 
     def step(self, x):
         self.n += 1
-        delta = x - self.mean
-        self.mean += delta/self.n
-        self.m2 += delta*(x - self.mean)
+        delta_ = x - self.mean
+        self.mean += delta_/self.n
+        self.m2 += delta_*(x - self.mean)
 
     def finalize(self):
         if self.entire_pop:
@@ -1049,7 +1160,7 @@ def indices_in_shape(shape):
     if isinstance(shape, int):
         shape = (shape,)
 
-    if len(shape) == 0:
+    if not shape:
         yield ()
     elif len(shape) == 1:
         for i in range(0, shape[0]):
@@ -1064,7 +1175,7 @@ def indices_in_shape(shape):
 def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
     """n may be a sequence, in which case length must be None."""
     if length is None:
-        if len(n) == 0:
+        if not n:
             yield ()
             return
 
@@ -1087,23 +1198,24 @@ def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
             yield my_part + base
 
 
-def generate_decreasing_nonnegative_tuples_summing_to(n, length, min=0, max=None):
+def generate_decreasing_nonnegative_tuples_summing_to(
+        n, length, min_value=0, max_value=None):
     if length == 0:
         yield ()
     elif length == 1:
-        if n <= max:
-            #print "MX", n, max
+        if n <= max_value:
+            #print "MX", n, max_value
             yield (n,)
         else:
             return
     else:
-        if max is None or n < max:
-            max = n
+        if max_value is None or n < max_value:
+            max_value = n
 
-        for i in range(min, max+1):
+        for i in range(min_value, max_value+1):
             #print "SIG", sig, i
             for remainder in generate_decreasing_nonnegative_tuples_summing_to(
-                    n-i, length-1, min, i):
+                    n-i, length-1, min_value, i):
                 yield (i,) + remainder
 
 
@@ -1163,29 +1275,29 @@ def generate_all_integer_tuples(length, least_abs=0):
 
 
 def generate_permutations(original):
-    """Generate all permutations of the list `original'.
+    """Generate all permutations of the list *original*.
 
     Nicked from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/252178
     """
     if len(original) <= 1:
         yield original
     else:
-        for perm in generate_permutations(original[1:]):
-            for i in range(len(perm)+1):
+        for perm_ in generate_permutations(original[1:]):
+            for i in range(len(perm_)+1):
                 #nb str[0:1] works in both string and list contexts
-                yield perm[:i] + original[0:1] + perm[i:]
+                yield perm_[:i] + original[0:1] + perm_[i:]
 
 
 def generate_unique_permutations(original):
-    """Generate all unique permutations of the list `original'.
+    """Generate all unique permutations of the list *original*.
     """
 
     had_those = set()
 
-    for perm in generate_permutations(original):
-        if perm not in had_those:
-            had_those.add(perm)
-            yield perm
+    for perm_ in generate_permutations(original):
+        if perm_ not in had_those:
+            had_those.add(perm_)
+            yield perm_
 
 
 def enumerate_basic_directions(dimensions):
@@ -1205,12 +1317,18 @@ def get_read_from_map_from_permutation(original, permuted):
     Requires that the permutation can be inferred from
     C{original} and C{permuted}.
 
-    >>> for p1 in generate_permutations(range(5)):
-    ...     for p2 in generate_permutations(range(5)):
-    ...         rfm = get_read_from_map_from_permutation(p1, p2)
-    ...         p2a = [p1[rfm[i]] for i in range(len(p1))]
-    ...         assert p2 == p2a
+    .. doctest ::
+
+        >>> for p1 in generate_permutations(range(5)):
+        ...     for p2 in generate_permutations(range(5)):
+        ...         rfm = get_read_from_map_from_permutation(p1, p2)
+        ...         p2a = [p1[rfm[i]] for i in range(len(p1))]
+        ...         assert p2 == p2a
     """
+    from warnings import warn
+    warn("get_read_from_map_from_permutation is deprecated and will be "
+            "removed in 2019", DeprecationWarning, stacklevel=2)
+
     assert len(original) == len(permuted)
     where_in_original = dict(
             (original[i], i) for i in range(len(original)))
@@ -1226,14 +1344,20 @@ def get_write_to_map_from_permutation(original, permuted):
     Requires that the permutation can be inferred from
     C{original} and C{permuted}.
 
-    >>> for p1 in generate_permutations(range(5)):
-    ...     for p2 in generate_permutations(range(5)):
-    ...         wtm = get_write_to_map_from_permutation(p1, p2)
-    ...         p2a = [0] * len(p2)
-    ...         for i, oi in enumerate(p1):
-    ...             p2a[wtm[i]] = oi
-    ...         assert p2 == p2a
+    .. doctest::
+
+        >>> for p1 in generate_permutations(range(5)):
+        ...     for p2 in generate_permutations(range(5)):
+        ...         wtm = get_write_to_map_from_permutation(p1, p2)
+        ...         p2a = [0] * len(p2)
+        ...         for i, oi in enumerate(p1):
+        ...             p2a[wtm[i]] = oi
+        ...         assert p2 == p2a
     """
+    from warnings import warn
+    warn("get_write_to_map_from_permutation is deprecated and will be "
+            "removed in 2019", DeprecationWarning, stacklevel=2)
+
     assert len(original) == len(permuted)
 
     where_in_permuted = dict(
@@ -1247,9 +1371,11 @@ def get_write_to_map_from_permutation(original, permuted):
 
 # {{{ graph algorithms
 
-def a_star(initial_state, goal_state, neighbor_map,
+def a_star(  # pylint: disable=too-many-locals
+        initial_state, goal_state, neighbor_map,
         estimate_remaining_cost=None,
-        get_step_cost=lambda x, y: 1):
+        get_step_cost=lambda x, y: 1
+        ):
     """
     With the default cost and heuristic, this amounts to Dijkstra's algorithm.
     """
@@ -1257,7 +1383,7 @@ def a_star(initial_state, goal_state, neighbor_map,
     from heapq import heappop, heappush
 
     if estimate_remaining_cost is None:
-        def estimate_remaining_cost(x):
+        def estimate_remaining_cost(x):  # pylint: disable=function-redefined
             if x != goal_state:
                 return 1
             else:
@@ -1278,7 +1404,7 @@ def a_star(initial_state, goal_state, neighbor_map,
     queue = [(init_remcost, AStarNode(initial_state, parent=None, path_cost=0))]
     visited_states = set()
 
-    while len(queue):
+    while queue:
         _, top = heappop(queue)
         visited_states.add(top.state)
 
@@ -1314,7 +1440,13 @@ def a_star(initial_state, goal_state, neighbor_map,
 # {{{ table formatting
 
 class Table:
-    """An ASCII table generator."""
+    """An ASCII table generator.
+
+    .. automethod:: add_row
+    .. automethod:: __str__
+    .. automethod:: latex
+    """
+
     def __init__(self):
         self.rows = []
 
@@ -1326,15 +1458,17 @@ class Table:
         col_widths = [max(len(row[i]) for row in self.rows)
                       for i in range(columns)]
 
-        lines = [
-            "|".join([cell.ljust(col_width)
-                      for cell, col_width in zip(row, col_widths)])
+        lines = [" | ".join([cell.ljust(col_width)
+            for cell, col_width in zip(row, col_widths)])
             for row in self.rows]
-        lines[1:1] = ["+".join("-"*col_width
-                              for col_width in col_widths)]
+        lines[1:1] = ["+".join("-" * (col_width + 1 + (i > 0))
+            for i, col_width in enumerate(col_widths))]
+
         return "\n".join(lines)
 
-    def latex(self, skip_lines=0, hline_after=[]):
+    def latex(self, skip_lines=0, hline_after=None):
+        if hline_after is None:
+            hline_after = []
         lines = []
         for row_nr, row in list(enumerate(self.rows))[skip_lines:]:
             lines.append(" & ".join(row)+r" \\")
@@ -1348,7 +1482,8 @@ class Table:
 
 # {{{ histogram formatting
 
-def string_histogram(iterable, min_value=None, max_value=None,
+def string_histogram(  # pylint: disable=too-many-arguments,too-many-locals
+        iterable, min_value=None, max_value=None,
         bin_count=20, width=70, bin_starts=None, use_unicode=True):
     if bin_starts is None:
         if min_value is None or max_value is None:
@@ -1370,7 +1505,7 @@ def string_histogram(iterable, min_value=None, max_value=None,
             bin_nr = bisect(bin_starts, value)-1
             try:
                 bins[bin_nr] += 1
-            except:
+            except Exception:
                 print(value, bin_nr, bin_starts)
                 raise
 
@@ -1402,13 +1537,13 @@ def string_histogram(iterable, min_value=None, max_value=None,
 
 def word_wrap(text, width, wrap_using="\n"):
     # http://code.activestate.com/recipes/148061-one-liner-word-wrap-function/
-    """
+    r"""
     A word-wrap function that preserves existing line breaks
     and most spaces in the text. Expects that existing line
-    breaks are posix newlines (\n).
+    breaks are posix newlines (``\n``).
     """
     space_or_break = [" ", wrap_using]
-    return reduce(lambda line, word, width=width: '%s%s%s' %
+    return reduce(lambda line, word: '%s%s%s' %
             (line,
                 space_or_break[(len(line)-line.rfind('\n')-1
                     + len(word.split('\n', 1)[0])
@@ -1434,7 +1569,11 @@ class CPyUserInterface(object):
     class Parameters(Record):
         pass
 
-    def __init__(self, variables, constants={}, doc={}):
+    def __init__(self, variables, constants=None, doc=None):
+        if constants is None:
+            constants = {}
+        if doc is None:
+            doc = {}
         self.variables = variables
         self.constants = constants
         self.doc = doc
@@ -1462,16 +1601,14 @@ class CPyUserInterface(object):
                 print("    %s" % self.doc[c])
 
     def gather(self, argv=None):
-        import sys
-
         if argv is None:
             argv = sys.argv
 
         if len(argv) == 1 or (
-                ("-h" in argv) or
-                ("help" in argv) or
-                ("-help" in argv) or
-                ("--help" in argv)):
+                ("-h" in argv)
+                or ("help" in argv)
+                or ("-help" in argv)
+                or ("--help" in argv)):
             self.show_usage(argv[0])
             sys.exit(2)
 
@@ -1523,17 +1660,19 @@ class MovedFunctionDeprecationWrapper:
 
 class StderrToStdout(object):
     def __enter__(self):
-        import sys
+        # pylint: disable=attribute-defined-outside-init
         self.stderr_backup = sys.stderr
         sys.stderr = sys.stdout
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        import sys
         sys.stderr = self.stderr_backup
         del self.stderr_backup
 
 
-def typedump(val, max_seq=5, special_handlers={}):
+def typedump(val, max_seq=5, special_handlers=None):
+    if special_handlers is None:
+        special_handlers = {}
+
     try:
         hdlr = special_handlers[type(val)]
     except KeyError:
@@ -1582,7 +1721,7 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
     if "EDITOR" in os.environ:
         from subprocess import Popen
         p = Popen([os.environ["EDITOR"], full_name])
-        os.waitpid(p.pid, 0)[1]
+        os.waitpid(p.pid, 0)
     else:
         print("(Set the EDITOR environment variable to be "
                 "dropped directly into an editor next time.)")
@@ -1600,7 +1739,15 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
 
 # {{{ progress bars
 
-class ProgressBar:
+class ProgressBar(object):  # pylint: disable=too-many-instance-attributes
+    """
+    .. automethod:: draw
+    .. automethod:: progress
+    .. automethod:: set_progress
+    .. automethod:: finished
+    .. automethod:: __enter__
+    .. automethod:: __exit__
+    """
     def __init__(self, descr, total, initial=0, length=40):
         import time
         self.description = descr
@@ -1640,7 +1787,6 @@ class ProgressBar:
             else:
                 eta_str = "?"
 
-            import sys
             sys.stderr.write("%-20s [%s] ETA %s\r" % (
                 self.description,
                 squares*"#"+(self.length-squares)*" ",
@@ -1657,7 +1803,6 @@ class ProgressBar:
         self.draw()
 
     def finished(self):
-        import sys
         self.set_progress(self.total)
         sys.stderr.write("\n")
 
@@ -1679,7 +1824,6 @@ def assert_not_a_file(name):
 
 
 def add_python_path_relative_to_script(rel_path):
-    import sys
     from os.path import dirname, join, abspath
 
     script_name = sys.argv[0]
@@ -1750,18 +1894,49 @@ def generate_unique_names(prefix):
         try_num += 1
 
 
+def generate_numbered_unique_names(prefix, num=None):
+    if num is None:
+        yield (0, prefix)
+        num = 0
+
+    while True:
+        name = "%s_%d" % (prefix, num)
+        num += 1
+        yield (num, name)
+
+
 generate_unique_possibilities = MovedFunctionDeprecationWrapper(
         generate_unique_names)
 
 
-class UniqueNameGenerator:
-    def __init__(self, existing_names=set(), forced_prefix=""):
+class UniqueNameGenerator(object):
+    """
+    .. automethod:: is_name_conflicting
+    .. automethod:: add_name
+    .. automethod:: add_names
+    .. automethod:: __call__
+    """
+    def __init__(self, existing_names=None, forced_prefix=""):
+        if existing_names is None:
+            existing_names = set()
+
         self.existing_names = existing_names.copy()
         self.forced_prefix = forced_prefix
+        self.prefix_to_counter = {}
 
     def is_name_conflicting(self, name):
         return name in self.existing_names
 
+    def _name_added(self, name):
+        """Callback to alert subclasses when a name has been added.
+
+        .. note::
+
+            This will not get called for the names in the *existing_names*
+            argument to :meth:`__init__`.
+        """
+        pass
+
     def add_name(self, name):
         if self.is_name_conflicting(name):
             raise ValueError("name '%s' conflicts with existing names")
@@ -1769,6 +1944,7 @@ class UniqueNameGenerator:
             raise ValueError("name '%s' does not start with required prefix")
 
         self.existing_names.add(name)
+        self._name_added(name)
 
     def add_names(self, names):
         for name in names:
@@ -1777,18 +1953,396 @@ class UniqueNameGenerator:
     def __call__(self, based_on="id"):
         based_on = self.forced_prefix + based_on
 
-        for var_name in generate_unique_names(based_on):
+        counter = self.prefix_to_counter.get(based_on, None)
+
+        for counter, var_name in generate_numbered_unique_names(based_on, counter):
             if not self.is_name_conflicting(var_name):
                 break
 
-        var_name = intern(var_name)
+        self.prefix_to_counter[based_on] = counter
+
+        var_name = intern(var_name)  # pylint: disable=undefined-loop-variable
 
         self.existing_names.add(var_name)
+        self._name_added(var_name)
         return var_name
 
 # }}}
 
 
+# {{{ recursion limit
+
+class MinRecursionLimit(object):
+    def __init__(self, min_rec_limit):
+        self.min_rec_limit = min_rec_limit
+
+    def __enter__(self):
+        # pylint: disable=attribute-defined-outside-init
+
+        self.prev_recursion_limit = sys.getrecursionlimit()
+        new_limit = max(self.prev_recursion_limit, self.min_rec_limit)
+        sys.setrecursionlimit(new_limit)
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        # Deep recursion can produce deeply nested data structures
+        # (or long chains of to-be gc'd generators) that cannot be
+        # undergo garbage collection with a lower recursion limit.
+        #
+        # As a result, it doesn't seem possible to lower the recursion limit
+        # again after it has been raised without causing reliability issues.
+        #
+        # See https://gitlab.tiker.net/inducer/sumpy/issues/31 for
+        # context.
+
+        pass
+
+# }}}
+
+
+# {{{ download from web if not present
+
+def download_from_web_if_not_present(url, local_name=None):
+    """
+    .. versionadded:: 2017.5
+    """
+
+    from os.path import basename, exists
+    if local_name is None:
+        local_name = basename(url)
+
+    if not exists(local_name):
+        from six.moves.urllib.request import urlopen
+        with urlopen(url) as inf:
+            contents = inf.read()
+
+            with open(local_name, "wb") as outf:
+                outf.write(contents)
+
+# }}}
+
+
+# {{{ find git revisions
+
+def find_git_revision(tree_root):  # pylint: disable=too-many-locals
+    # Keep this routine self-contained so that it can be copy-pasted into
+    # setup.py.
+
+    from os.path import join, exists, abspath
+    tree_root = abspath(tree_root)
+
+    if not exists(join(tree_root, ".git")):
+        return None
+
+    # construct minimal environment
+    # stolen from
+    # https://github.com/numpy/numpy/blob/055ce3e90b50b5f9ef8cf1b8641c42e391f10735/setup.py#L70-L92
+    import os
+    env = {}
+    for k in ['SYSTEMROOT', 'PATH', 'HOME']:
+        v = os.environ.get(k)
+        if v is not None:
+            env[k] = v
+    # LANGUAGE is used on win32
+    env['LANGUAGE'] = 'C'
+    env['LANG'] = 'C'
+    env['LC_ALL'] = 'C'
+
+    from subprocess import Popen, PIPE, STDOUT
+    p = Popen(["git", "rev-parse", "HEAD"], shell=False,
+              stdin=PIPE, stdout=PIPE, stderr=STDOUT, close_fds=True,
+              cwd=tree_root, env=env)
+    (git_rev, _) = p.communicate()
+
+    if sys.version_info >= (3,):
+        git_rev = git_rev.decode()
+
+    git_rev = git_rev.rstrip()
+
+    retcode = p.returncode
+    assert retcode is not None
+    if retcode != 0:
+        from warnings import warn
+        warn("unable to find git revision")
+        return None
+
+    return git_rev
+
+
+def find_module_git_revision(module_file, n_levels_up):
+    from os.path import dirname, join
+    tree_root = join(*([dirname(module_file)] + [".." * n_levels_up]))
+
+    return find_git_revision(tree_root)
+
+# }}}
+
+
+# {{{ create a reshaped view of a numpy array
+
+def reshaped_view(a, newshape):
+    """ Create a new view object with shape ``newshape`` without copying the data of
+    ``a``. This function is different from ``numpy.reshape`` by raising an
+    exception when data copy is necessary.
+
+    :arg a: a :class:`numpy.ndarray` object.
+    :arg newshape: an ``int`` object or a tuple of ``int`` objects.
+
+    .. versionadded:: 2018.4
+    """
+
+    newview = a.view()
+    newview.shape = newshape
+    return newview
+
+# }}}
+
+
+# {{{ process timer
+
+SUPPORTS_PROCESS_TIME = (sys.version_info >= (3, 3))
+
+
+class ProcessTimer(object):
+    """Measures elapsed wall time and process time.
+
+    .. automethod:: __enter__
+    .. automethod:: __exit__
+    .. automethod:: done
+
+    Timing data attributes:
+
+    .. attribute:: wall_elapsed
+    .. attribute:: process_elapsed
+
+        Only available in Python 3.3+.
+
+    .. versionadded:: 2018.5
+    """
+
+    def __init__(self):
+        import time
+        if SUPPORTS_PROCESS_TIME:
+            self.perf_counter_start = time.perf_counter()
+            self.process_time_start = time.process_time()
+
+        else:
+            import timeit
+            self.time_start = timeit.default_timer()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.done()
+
+    def done(self):
+        # pylint: disable=attribute-defined-outside-init
+
+        import time
+        if SUPPORTS_PROCESS_TIME:
+            self.wall_elapsed = time.perf_counter() - self.perf_counter_start
+            self.process_elapsed = time.process_time() - self.process_time_start
+
+        else:
+            import timeit
+            self.wall_elapsed = timeit.default_timer() - self.time_start
+            self.process_elapsed = None
+
+# }}}
+
+
+# {{{ log utilities
+
+class ProcessLogger(object):  # pylint: disable=too-many-instance-attributes
+    """Logs the completion time of a (presumably) lengthy process to :mod:`logging`.
+    Only uses a high log level if the process took perceptible time.
+
+    .. automethod:: __init__
+    .. automethod:: done
+    .. automethod:: __enter__
+    .. automethod:: __exit__
+    """
+
+    default_noisy_level = logging.INFO
+
+    def __init__(  # pylint: disable=too-many-arguments
+            self, logger, description,
+            silent_level=None, noisy_level=None, long_threshold_seconds=None):
+        self.logger = logger
+        self.description = description
+        self.silent_level = silent_level or logging.DEBUG
+        self.noisy_level = noisy_level or self.default_noisy_level
+        self.long_threshold_seconds = (
+                # 0 is a valid value that should override the default
+                0.3 if long_threshold_seconds is None else long_threshold_seconds)
+
+        self.logger.log(self.silent_level, "%s: start", self.description)
+        self.is_done = False
+
+        import threading
+        self.late_start_log_thread = threading.Thread(target=self._log_start_if_long)
+
+        # Do not delay interpreter exit if thread not finished.
+        self.late_start_log_thread.daemon = True
+
+        # https://github.com/firedrakeproject/firedrake/issues/1422
+        # Starting a thread may irrecoverably break various environments,
+        # e.g. MPI.
+        #
+        # Since the late-start logging is an optional 'quality-of-life'
+        # feature for interactive use, do not do it unless there is (weak)
+        # evidence of interactive use.
+        import sys
+        if sys.stdin is None:
+            # Can happen, e.g., if pudb is controlling the console.
+            use_late_start_logging = False
+        else:
+            use_late_start_logging = sys.stdin.isatty()
+
+        import os
+        if os.environ.get("PYTOOLS_LOG_NO_THREADS", ""):
+            use_late_start_logging = False
+
+        if use_late_start_logging:
+            try:
+                self.late_start_log_thread.start()
+            except RuntimeError:
+                # https://github.com/firedrakeproject/firedrake/issues/1422
+                #
+                # Starting a thread may fail in various environments, e.g. MPI.
+                # Since the late-start logging is an optional 'quality-of-life'
+                # feature for interactive use, tolerate failures of it without
+                # warning.
+                pass
+
+        self.timer = ProcessTimer()
+
+    def _log_start_if_long(self):
+        from time import sleep
+
+        sleep_duration = 10*self.long_threshold_seconds
+        sleep(sleep_duration)
+
+        if not self.is_done:
+            self.logger.log(
+                    self.noisy_level, "%s: started %.gs ago",
+                    self.description,
+                    sleep_duration)
+
+    def done(  # pylint: disable=keyword-arg-before-vararg
+            self, extra_msg=None, *extra_fmt_args):
+        self.timer.done()
+        self.is_done = True
+
+        wall_elapsed = self.timer.wall_elapsed
+        process_elapsed = self.timer.process_elapsed
+
+        completion_level = (
+                self.noisy_level
+                if wall_elapsed > self.long_threshold_seconds
+                else self.silent_level)
+
+        if process_elapsed is not None:
+            msg = "%s: completed (%.2fs wall, %.1fx CPU)"
+            fmt_args = [self.description, wall_elapsed, process_elapsed/wall_elapsed]
+        else:
+            msg = "%s: completed (%f.2s wall)"
+            fmt_args = [self.description, wall_elapsed]
+
+        if extra_msg:
+            msg += ": " + extra_msg
+            fmt_args.extend(extra_fmt_args)
+
+        self.logger.log(completion_level, msg, *fmt_args)
+
+    def __enter__(self):
+        pass
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.done()
+
+
+class DebugProcessLogger(ProcessLogger):
+    default_noisy_level = logging.DEBUG
+
+
+class log_process(object):  # noqa: N801
+    """A decorator that uses :class:`ProcessLogger` to log data about calls
+    to the wrapped function.
+    """
+
+    def __init__(self, logger, description=None):
+        self.logger = logger
+        self.description = description
+
+    def __call__(self, wrapped):
+        def wrapper(*args, **kwargs):
+            with ProcessLogger(
+                    self.logger,
+                    self.description or wrapped.__name__):
+                return wrapped(*args, **kwargs)
+
+        from functools import update_wrapper
+        new_wrapper = update_wrapper(wrapper, wrapped)
+
+        return new_wrapper
+
+# }}}
+
+
+# {{{ sorting in natural order
+
+def natorder(item):
+    """Return a key for natural order string comparison.
+
+    See :func:`natsorted`.
+
+    .. versionadded:: 2020.1
+    """
+    import re
+    result = []
+    for (int_val, string_val) in re.findall(r"(\d+)|(\D+)", item):
+        if int_val:
+            result.append(int(int_val))
+            # Tie-breaker in case of leading zeros in *int_val*.  Longer values
+            # compare smaller to preserve order of numbers in decimal notation,
+            # e.g., "1.001" < "1.01"
+            # (cf. https://github.com/sourcefrog/natsort)
+            result.append(-len(int_val))
+        else:
+            result.append(string_val)
+    return result
+
+
+def natsorted(iterable, key=None, reverse=False):
+    """Sort using natural order [1]_, as opposed to lexicographic order.
+
+    Example::
+
+        >>> sorted(["_10", "_1", "_9"]) == ["_1", "_10", "_9"]
+        True
+        >>> natsorted(["_10", "_1", "_9"]) == ["_1", "_9", "_10"]
+        True
+
+    :arg iterable: an iterable to be sorted. It must only have strings, unless
+        *key* is specified.
+    :arg key: if provided, a key function that returns strings for ordering
+        using natural order.
+    :arg reverse: if *True*, sorts in descending order.
+
+    :returns: a sorted list
+
+    .. [1] https://en.wikipedia.org/wiki/Natural_sort_order
+
+    .. versionadded:: 2020.1
+    """
+    if key is None:
+        key = lambda x: x
+    return sorted(iterable, key=lambda y: natorder(key(y)), reverse=reverse)
+
+# }}}
+
+
 def _test():
     import doctest
     doctest.testmod()
